
%if 0

UNFILL - Free the UMBs allocated by my UMBFILL
Written by C. Masloch, 2008. Public Domain.
Updated 2022.

UNFILL identifies UMBs filled by UMBFILL by first verifying if the MCB name is "UMBFILL",0.
If it's not, this is no block filled by UMBFILL. Then it's checked wether the MCB owner value
equals the actual block. If it does, then this a block filled by UMBFILL. If it doesn't, it's
last checked wether the MCB owner value equals 0008h or another block that points to itself
and also has the MCB name of "UMBFILL",0. If either condition matches, UMBFILL's PSP was
loaded here because UMBFILL was loaded with LH (and the UNFILL will free this block as well).

Update:	Now we default to accept only blocks owned by themselves, with
	 the matching MCB name, and either memory block size zero, or with the
	 first paragraph of the memory block matching an additional ID. This
	 is simply the MCB name repeated twice.
	PSP blocks are identified as either owner = 0008h, or the owner being
	 another UMBFILL block that matches the first (normal) check, including
	 the additional ID in the first data paragraph.
	Passing an exclamation point to UNFILL on the command line makes it
	 ignore the additional ID, allowing to free blocks allocated by older
	 UMBFILL releases as well.

Note:	We actually do two passes over all MCBs. In the first pass we identify
	 potential PSP MCBs and free them. In the second pass we identify and
	 free ordinary blocks. This is needed because a PSP block may refer to
	 an ordinary block, so to avoid use-after-free we should handle PSP
	 blocks in the first pass, where ordinary blocks are still in use.

%endif


		cpu 8086
		org 100h

start:
		mov si, 81h
.clnext:
		lodsb
		cmp al, '!'			; flag ?
		je .clgot			; yes -->
		cmp al, 13			; end ?
		jne .clnext			; not yet -->
		jmp .cldone			; done -->

.clgot:
		inc byte [ignore_id]		; set the flag
.cldone:

	mov ax, 5802h
	int 21h
	xor ah, ah		; some "DOS" only return al
	mov word [ restored.umblink ], ax	; save UMB link

	; mov ax, 5800h
	; int 21h
	; mov word [ restored.strategy ], ax	; save strategy

	mov dx, i23
	mov ax, 2523h
	int 21h			; set Int23 so that we restore link/strategy on Ctrl-C as well
	mov dx, i24
	mov ax, 2524h
	int 21h			; set Int24 so all hard errors are changed to the soft error

		mov ax, 5803h			; set UMB link state
		mov bx, 1			; enabled
		int 21h
	jc error

		xor dx, dx			; dh = PSP flag. 0= scan for the UMBFILL PSP (special handling)
						; dl = End flag. 0= Current MCB is an 'M' MCB, 1= It's the 'Z' MCB
		xor cx, cx			; set UMB counter to 0

getfirstmcb:
		mov ah, 52h
		int 21h				; get DOS list of lists
		mov bx, [ es:bx - 2 ]		; get first MCB

		push cs
		pop es

nextmcb:		; scan in next MCB
			; bx = Next MCB, dx = search flags, cx = number of already freed UMBs
		push ds
		mov ds, bx

		cmp byte [ 0000h ], 'M'		; not last MCB ?
		je .valid			; yes -->

		inc dl				; indicate last MCB
		cmp byte [ 0000h ], 'Z'		; valid MCB ?
		jne error			; no -->

.valid:
		cmp bx, 9FFFh			; is it an UMB at all ?
		jb .next			; no, just walk to next -->

		call checkmcb
		jc .next			; doesn't match -->

		inc cx				; found matching, increase counter

		push es
		mov ax, ds
		inc ax
		mov es, ax			; must pass block segment to DOS
		mov ah, 49h			; free memory
		int 21h
		pop es

.next:
		add bx, word [ 0003h ]		; add size of this block
		inc bx				; add one to get next PSP
		pop ds
		cmp dl, 0			; last MCB was 'Z' ?
		je nextmcb			; no, continue until 'Z' MCB hit
		xor dl, dl			; reset to 0
		cmp dh, 1			; Only PSP search completed ? (if below, it's zero)
		inc dh				; indicate to search normal UMBFILL blocks now
						; (inc doesn't affect CF)
		jb getfirstmcb			; yes, start normal search from first MCB -->

checkifany:
		jcxz error			; no matching UMBs found at all ? -->

		mov dx, msg.done
		mov ah, 09h
		int 21h

	xor ax, ax
	jmp terminate


error:
		mov dx, msg.noumbs
		mov ah, 09h
		int 21h
	mov al, 1

terminate:
	call restorestate
	mov ah, 4Ch
	int 21h


i24:
	mov al, 3			; always return fail, to handle the error as a soft one
	iret


%define __JMP_REL8 0EBh
%define __REL__(address,fixup) ((address)-(fixup))
%xdefine __REL16__(address) __REL__(address,$+2); for __JMP_REL16 (dw)

i23:
	mov word [ cs:$ ], (__JMP_REL8|__REL16__(.return)<<8)	; don't reenter
	call restorestate
.return:
	stc				; always abort program (what default DOS handler also does)
	retf


		; Restore modified DOS data
		;
		; CHG:	-
		; USE:	Int21
restorestate:
	push ax
	push bx
	; mov bx, word [ cs:restored.strategy ]
	; mov ax, 5801h
	; int 21h			; restore strategy

	mov bx, word [ cs:restored.umblink ]
	mov ax, 5803h
	int 21h				; restore UMB link
	pop bx
	pop ax
	retn


checkmcb:		; check if MCB is an UMBFILL MCB
			; es = cs, ds = checked MCB
			; dh = 0 if UMBFILL PSP special check, = 1 if normal check
			;
			; Returns CY if no UMBFILL MCB, NC else
			;
			; Uses di, si and ax
		cmp word [ 0001h ], byte 0000h	; already free ?
		je .no				; yes, then no match -->

		mov di, msg.mcbname
		mov si, 0008h

		push cx
		mov cx, 4			; 4 words
		 push di
		repe cmpsw			; does the MCB name match ?
		 pop di
		pop cx
		jne .no				; no -->

		cmp dh, 0
		je .psp
.normal:
		mov ax, ds
		inc ax
		cmp ax, word [ 0001h ]		; owner references block itself ?
		jne .no				; no -->
	cmp byte [cs:ignore_id], 1		; ignore id ?
	je .yes					; yes --> (ZR, NC)
	cmp word [0003h], 0			; is it an empty block ?
	je .yes					; yes --> (ZR, NC)
	push cx
	mov cx, 4				; 4 words
	 push di
	repe cmpsw				; check first signature half
	 pop di
	jne .normal_decide			; if not it, skip next half -->
	mov cl, 4				; 4 words (ch still zero)
	repe cmpsw				; check second signature half
.normal_decide:
	pop cx
	jne .no					; does not match ID -->
	retn					; else, match. flags ZR, NC

.psp:
		call .normal
		jnc .no				; mustn't point to itself -->
		mov ax, word [ 0001h ]
		cmp ax, byte 0008h		; marked as owned by DOS ?
		je .yes				; yes, matches -->

		push ds
		push dx
		dec ax
		mov ds, ax			; get the MCB of the owner block
		inc dh				; indicate to check normal now
		call checkmcb			; check if it's ok
		pop dx
		pop ds
						; leave CF
.yes:
		retn

.no:
		stc
		retn


ignore_id:	db 0
	align 2, db 0
restored:
.umblink:	dw 0
; .strategy:	dw 0


msg:
	align 2, db 0
.mcbname:	db "UMBFILL"
		times 8 - ($ - .mcbname) db 0
.noumbs:	db "No UMBFILL UMBs found",13,10,36
.done:		db "All UMBFILL UMBs deallocated",13,10,36
